
/**
 * Drupal initializer.
 * Launch as behavior and pull variables from config.
 */

  // Prevent multiple inits in modules that re-trigger the document context.
let ed11yOnce;
let ed11yInitialized;

Drupal.behaviors.editoria11y = {
  attach: function (context, settings) {

    if (context === document && !ed11yOnce && CSS.supports('selector(:is(body))')) {

      /**
       * Initiate library
       *
       * */
        // "Supports" selector drops old browsers.

      let options = {};

      // todo: store dismissalKeys for PDFs in page results, and check dismissals table for page level matches on load.

      options.checkRoots = drupalSettings.editoria11y.content_root ? drupalSettings.editoria11y.content_root : false;
      options.ignoreElements = !!drupalSettings.editoria11y.ignore_elements ? `#toolbar-administration *, ${drupalSettings.editoria11y.ignore_elements}` : '#toolbar-administration *';
      options.panelNoCover = !!drupalSettings.editoria11y.panel_no_cover ?
        drupalSettings.editoria11y.panel_no_cover :
        '#klaro_toggle_dialog'; // todo expose param

      options.ignoreAllIfAbsent = !!drupalSettings.editoria11y.ignore_all_if_absent ? drupalSettings.editoria11y.ignore_all_if_absent : false;
      if (document.URL.indexOf('mode=same_page_preview') > -1) {
        // Todo postpone: Same page preview refreshes its frame.
        // We would need an inside/outside dance like WP
        // Todo postpone: offset Editoria11y toggle by width of SPP iframe?
        ed11yOnce = true;
        return;
      } else if (drupalSettings.path.currentPathIsAdmin) {
        if (document.querySelector('.layout-builder-form')) {
          options.inlineAlerts = false;
        } else if (!drupalSettings.editor) {
          // Ed11y will init later if a behavior brings in something editable.
          return;
        } else {
          // Editable content is present.
          // todo postpone: expose these as parameters?.
          options.checkRoots = '[contentEditable="true"]';
          options.ignoreElements += ', [hidden], [style*="display: none"], [style*="display:none"], [hidden] *, [style*="display: none"] *, [style*="display:none"] *';
          options.ignoreAllIfAbsent = '[contentEditable="true"]';
        }
      }
      ed11yOnce = true;

      // todo: parameter for assertiveness or user pref on live editing?

      let urlParams = new URLSearchParams(window.location.search);

      let lang = drupalSettings.editoria11y.lang ? drupalSettings.editoria11y.lang : 'en';

      if (lang !== 'en') {
        lang = 'dynamic';
        ed11yLang.dynamic = ed11yLangDrupal;
      }

      let ed11yAlertMode = drupalSettings.editoria11y.assertiveness ? drupalSettings.editoria11y.assertiveness : 'assertive';
      // If assertiveness is "smart" we set it to assertive if the doc was recently changed.
      const now = new Date();
      if (drupalSettings.path.currentPathIsAdmin && drupalSettings.editor && drupalSettings.editoria11y.assertiveness !== 'polite') {
        ed11yAlertMode = 'active';
      }
      else if (
        urlParams.has('ed1ref') ||
        (ed11yAlertMode === 'smart' &&
          ((now / 1000) - drupalSettings.editoria11y.changed < 60)
        )
      ) {
        ed11yAlertMode = 'assertive';
      }


      options.lang = lang;
      options.ignoreByKey = {
        'p': 'table:not(.field-multiple-table) p',
        'h': '.filter-guidelines-item *, nav *, .block-local-tasks-block *, .tabledrag h4',
        // disable alt text tests on unspoken images
        'img': '[aria-hidden], [aria-hidden] img',
        // disable link text check on disabled links:
        'a': `[aria-hidden][tabindex], .block-local-tasks-block a, .filter-help > a, .contextual-region > nav a ${drupalSettings.path.currentPathIsAdmin ? ', a[target="_blank"]' : ''}`,
        // 'li': false,
        // 'blockquote': false,
        // 'iframe': false,
        // 'audio': false,
        // 'video': false,
        'table': '[role="presentation"], .tabledrag',
        // todo: report h4 and th issue in docroot/core/includes/theme.inc
      }
      options.alertMode = ed11yAlertMode;
      options.currentPage = drupalSettings.editoria11y.page_path;
      options.allowHide = !!drupalSettings.editoria11y.allow_hide;
      options.allowOK = !!drupalSettings.editoria11y.allow_ok;
      options.syncedDismissals = drupalSettings.editoria11y.dismissals;
      options.showDismissed = urlParams.has('ed1ref');
      // todo postpone: ignoreAllIfPresent
      options.preventCheckingIfPresent = !!drupalSettings.editoria11y.no_load ? drupalSettings.editoria11y.no_load : false;
      // todo postpone: preventCheckingIfAbsent
      options.linkStringsNewWindows = !!drupalSettings.editoria11y.link_strings_new_windows ?
        new RegExp (drupalSettings.editoria11y.link_strings_new_windows, 'g')
        : !!drupalSettings.editoria11y.ignore_link_strings ?
          new RegExp(drupalSettings.editoria11y.ignore_link_strings, 'g')
          : new RegExp ('(' + Drupal.t('download') + ')|(\\s' + Drupal.t('tab') + ')|(' + Drupal.t('window') + ')', 'g');
      options.linkIgnoreStrings = !!drupalSettings.editoria11y.ignore_link_strings ? new RegExp(drupalSettings.editoria11y.ignore_link_strings, 'g') : new RegExp('(' + Drupal.t('link is external') + ')|(' + Drupal.t('link sends email') + ')', 'g');
      options.linkIgnoreSelector = !!drupalSettings.editoria11y.link_ignore_selector ? drupalSettings.editoria11y.link_ignore_selector : false;
      options.hiddenHandlers = !!drupalSettings.editoria11y.hidden_handlers ? drupalSettings.editoria11y.hidden_handlers : '';
      options.theme = !!drupalSettings.editoria11y.theme ? drupalSettings.editoria11y.theme : 'sleekTheme';
      options.embeddedContent = !!drupalSettings.editoria11y.embedded_content_warning ? drupalSettings.editoria11y.embedded_content_warning : false;
      options.documentLinks = !!drupalSettings.editoria11y.download_links ? drupalSettings.editoria11y.download_links : `a[href$='.pdf'], a[href*='.pdf?'], a[href$='.doc'], a[href$='.docx'], a[href*='.doc?'], a[href*='.docx?'], a[href$='.ppt'], a[href$='.pptx'], a[href*='.ppt?'], a[href*='.pptx?'], a[href^='https://docs.google']`;
      options.buttonZIndex = 503;
      options.customTests = drupalSettings.editoria11y.custom_tests;
      options.cssUrls = !!drupalSettings.editoria11y.css_url ? [drupalSettings.editoria11y.css_url + '/library/css/editoria11y.css'] : false;

      /*let options = {
        // videoContent: 'youtube.com, vimeo.com, yuja.com, panopto.com',
        // audioContent: 'soundcloud.com, simplecast.com, podbean.com, buzzsprout.com, blubrry.com, transistor.fm, fusebox.fm, libsyn.com',
        // dataVizContent: 'datastudio.google.com, tableau',
        // twitterContent: 'twitter-timeline',
        ,
        editableContent: '[contentEditable="true"], #quickedit-entity-toolbar, .layout-builder-form',
        shadowComponents: drupalSettings.editoria11y.shadow_components ? drupalSettings.editoria11y.shadow_components : false,
        ,
      };*/

      if (typeof editoria11yOptionsOverride !== 'undefined' && typeof editoria11yOptions === 'function') {
        options = editoria11yOptions(options);
      }

      if (!!drupalSettings.editoria11y.view_reports) {
        options.reportsURL = drupalSettings.editoria11y.dashboard_url;
      }

      const ed11y = new Ed11y(options);

      ed11yInitialized = true;

      // Only needed if we have issues with the mutation observers.
      /*window.setTimeout(function() {
        let cks = Drupal.CKEditor5Instances;
        cks?.forEach(ck => {
          console.log(ck);
          ck.model.document.on( 'change:data', () => {
            console.log('change');
            window.setTimeout(() => {Ed11y.incrementalCheck();}, 100);
          } );
        })
      }, 1000);
      */

      /**
       * Initiate sync
       *
       * */

      let csrfToken = false;
      function getCsrfToken(action, data)
      {
        {
          fetch(`${drupalSettings.editoria11y.session_url}`, {
            method: "GET"
          })
            .then(res => res.text())
            .then(token => {
              csrfToken = token;
              postData(action, data).catch(err => console.error(err));
            })
            .catch(err => console.error(err));
        }
      }

      let postData = async function (action, data) {
        if (!csrfToken) {
          getCsrfToken(action, data);
        } else {
          let apiRoot = drupalSettings.editoria11y.api_url.replace('results/report','');
          let url = `${apiRoot}${action}`;
          fetch(url, {
            method: 'POST',
            headers: {
              'Content-Type': 'application/json',
              'X-CSRF-Token': csrfToken,
            },
            body: JSON.stringify(data),
          })
            .catch((error) => console.error('Error:', error))
        }
      }

      // Purge changed aliases & deleted pages.
      if (urlParams.has('ed1ref') && urlParams.get('ed1ref') !== drupalSettings.editoria11y.page_path) {
        let data = {
          page_path: urlParams.get('ed1ref'),
        };
        window.setTimeout(function() {
          postData('purge/page', data);
        },0,data);
      }

      let results = {};
      let oks = {};
      let total = 0;
      let extractResults = function () {
        results = {};
        oks = {};
        total = 0;
        Ed11y.results.forEach(result => {
          if (result.dismissalStatus !== "ok") {
            // log all items not marked as OK
            let testName = result.test;
            testName = Ed11y.M[testName].title;
            if (results[testName]) {
              results[testName] = parseInt(results[testName]) + 1;
              total++;
            } else {
              results[testName] = 1;
              total++;
            }
          }
          if (result.dismissalStatus === "ok") {
            if (!results[result.test]) {
              oks[result.test] = Ed11y.M[result.test].title;
            }
          }
        })
      }

      let sendResults = function () {
        window.setTimeout(function () {
          total = 0;
          extractResults();
          let url = window.location.pathname + window.location.search;
          url = url.length > 1000 ? url.substring(0, 1000) : url;
          let data = {
            page_title: drupalSettings.editoria11y.page_title,
            page_path: drupalSettings.editoria11y.page_path,
            entity_id: drupalSettings.editoria11y.entity_id,
            page_count: total,
            language: drupalSettings.editoria11y.lang,
            entity_type: drupalSettings.editoria11y.entity_type, // node or false
            route_name: drupalSettings.editoria11y.route_name, // e.g., entity.node.canonical or view.frontpage.page_1
            results: results,
            oks: oks,
            page_url: url,
          };
          postData('results/report', data);
          // Short timeout to let execution queue clear.
        }, 100)
      }

      let firstRun = true;
      if (drupalSettings.editoria11y.dismissals && drupalSettings.editoria11y.sync !== 'dismissals') {
        document.addEventListener('ed11yResults', function () {
          if (firstRun) {
            sendResults();
            firstRun = false;
          }
        });
      }

      let sendDismissal = function (detail) {
        if (!!detail) {
          let data = {};
          if (detail.dismissAction === 'reset') {
            data = {
              page_path: drupalSettings.editoria11y.page_path,
              language: drupalSettings.editoria11y.lang,
              route_name: drupalSettings.editoria11y.route_name,
              dismissal_status: 'reset', // ok, ignore or reset
            };
            window.setTimeout(function() {
              sendResults();
            },100);
          } else {
            data = {
              page_title: drupalSettings.editoria11y.page_title,
              page_path: drupalSettings.editoria11y.page_path,
              entity_id: drupalSettings.editoria11y.entity_id,
              language: drupalSettings.editoria11y.lang,
              entity_type: drupalSettings.editoria11y.entity_type, // node or false
              route_name: drupalSettings.editoria11y.route_name, // e.g., entity.node.canonical or view.frontpage.page_1
              result_name: Ed11y.M[detail.dismissTest].title, // which test is sending a result
              result_key: detail.dismissTest, // which test is sending a result
              element_id: detail.dismissKey, // some recognizable attribute of the item marked
              dismissal_status: detail.dismissAction, // ok, ignore or reset
            };
            if (detail.dismissAction === 'ok') {
              window.setTimeout(function() {
                sendResults();
              },100);
            }
          }
          postData('dismiss/' + detail.dismissAction, data);
        }
      }
      if (drupalSettings.editoria11y.dismissals) {
        document.addEventListener('ed11yDismissalUpdate', function (e) {
          sendDismissal(e.detail)}, false);
      }
    }

    if (ed11yInitialized) {
      // Recheck page about a second after every behavior.
      // Todo: global mutation watch instead or in addition?
      window.setTimeout(function () {
        Ed11y.forceFullCheck = true;
        Ed11y.incrementalCheck();
      }, 1000);
    }

  }
};
